// Adapted for LHAPDFv4 by Mike Whalley.
// Adapted for LHAPDFv5 by Craig Group & Mike Whalley.
// Fortran portability and string passing by Andy Buckley.

#include "LHAPDF/LHAPDF.h"
#include "LHAPDF/LHAPDFfw.h"

#include <cstring>
#include <iostream>
#include <stdexcept>
using namespace std;

namespace LHAPDF {


  string _getFileExtension(SetType type) {
    switch(type) {
    case LHPDF:
      return "LHpdf";
    case LHGRID:
      return "LHgrid";
    }
    // This should never happen: here to remove compiler warnings.
    throw runtime_error("Unknown file extension enum value in LHAPDF::_getFileExtension");
    return "";
  }

  string _mkStringFromFortran(char* fstring, size_t length) {
    // NB. "length" is the length of the target C string: the Fortran char
    // array must be one element larger, so that the trailing null can be 
    // appended, like so:
    fstring[length+1] = '\0';
    for (size_t i = length; i >= 0; --i) {
      //cout << i << ": `" << fstring[i] << "'" << endl;
      if (fstring[i] != ' ') break;
      //fstring[i] = '@'; //'\0';
      fstring[i] = '\0';
    }
    return string(fstring);
  }

  /////////////////


  void setVerbosity(Verbosity noiselevel) { 
    switch (noiselevel) {
    case SILENT:
      fsilent();
      break;
    case LOWKEY:
      flowkey();
      break;
    default:
      fdefaultverb();
    }
  }

  void setPDFPath(const string& path) { 
    char cpath[1000];
    strncpy(cpath, path.c_str(), 1000);
    fsetpdfpath(cpath, path.length());
  }


  ///////////////


  string prefixPath() {
    char str[1000];
    fgetprefixpath(str, 1000);
    return _mkStringFromFortran(str, 999);
  }


  string pdfsetsPath() {
    char str[1000];
    fgetdirpath(str, 1000);
    return _mkStringFromFortran(str, 999);
  }


  string pdfsetsIndexPath() {
    char str[1000];
    fgetindexpath(str, 1000);
    return _mkStringFromFortran(str, 999);
  }


  ///////////////


  void initPDFSet(const std::string& name, SetType type, int member) {
    initPDFSet(name + "." + _getFileExtension(type), member);
  }

  void initPDFSet(int nset, const std::string& name, SetType type, int member) {
    initPDFSet(nset, name + "." + _getFileExtension(type), member);
  }

  void initPDFSet(const std::string& filename, int member) {
    char cfilename[1000];
    strncpy(cfilename, filename.c_str(), 1000);
    if (filename.find("/") == string::npos) {
      finitpdfsetbyname(cfilename, filename.length());
    } else {
      finitpdfset(cfilename, filename.length());
    }
    usePDFMember(member);
  }

  void initPDFSet(int nset, const std::string& filename, int member) {
    char cfilename[1000];
    strncpy(cfilename, filename.c_str(), 1000);
    if (filename.find("/") == string::npos) {
      finitpdfsetbynamem(&nset, cfilename, filename.length());
    } else {
      finitpdfsetm(&nset, cfilename, filename.length());
    }
    usePDFMember(nset, member);
  }

  void usePDFMember(int member) {
    finitpdf(&member);
  }

  void usePDFMember(int nset, int member) {
    finitpdfm(&nset, &member);
  }


  ///////////////


  void initPDFByName(const string& name, SetType type, int memset) {
    initPDFSetByName(name, type);
    initPDF(memset);
  }
  void initPDFByName(int nset, const string& name, SetType type, int memset) {
    initPDFSetByName(nset, name, type);
    initPDF(nset, memset);
  }

  void initPDFByName(const string& filename, int memset) {
    initPDFSetByName(filename);
    initPDF(memset);
  }
  void initPDFByName(int nset, const string& filename, int memset) {
    initPDFSetByName(nset, filename);
    initPDF(nset, memset);
  }

  // void initPDFSet(const string& path) { 
  //   char cpath[1000];
  //   strncpy(cpath, path.c_str(), 1000);
  //   finitpdfset(cpath, path.length());
  // }
  // void initPDFSet(int nset, const string& path) {
  //   char cpath[1000];
  //   strncpy(cpath, path.c_str(), 1000);
  //   finitpdfsetm(&nset, cpath, path.length());
  // }

  void initPDFSetByName(const string& name, SetType type) {
    initPDFSetByName(name + "." + _getFileExtension(type));
  }
  void initPDFSetByName(int nset, const string& name, SetType type) {
    initPDFSetByName(nset, name + "." + _getFileExtension(type));
  }
  
  void initPDFSetByName(const string& filename) { 
    char cfilename[1000];
    strncpy(cfilename, filename.c_str(), 1000);
    finitpdfsetbyname(cfilename, filename.length());
  }
  void initPDFSetByName(int nset, const string& filename) {
    char cfilename[1000];
    strncpy(cfilename, filename.c_str(), 1000);
    finitpdfsetbynamem(&nset, cfilename, filename.length());
  }

  void initPDF(int memset) { 
    finitpdf(&memset);
  }
  void initPDF(int nset, int memset) {
    finitpdfm(&nset, &memset);
  }


  //////////////////


  vector<double> xfx(double x, double Q) {  
    vector<double> r(13);
    fevolvepdf(&x, &Q, &r[0]);
    return r;
  }
  vector<double> xfx(int nset, double x, double Q) {  
    vector<double> r(13);
    fevolvepdfm(&nset, &x, &Q, &r[0]);
    return r;
  }

  void xfx(double x, double Q, double* results) {
    fevolvepdf(&x, &Q, results);
  }
  void xfx(int nset, double x, double Q, double* results) {
    fevolvepdfm(&nset, &x, &Q, results);
  }

  double xfx(double x, double Q, int fl) {  
    vector<double> r(13);
    fevolvepdf(&x, &Q, &r[0]);
    return r[fl+6];
  }
  double xfx(int nset, double x, double Q, int fl) {  
    vector<double> r(13);
    fevolvepdfm(&nset, &x, &Q, &r[0]);
    return r[fl+6];
  }



  vector<double> xfxp(double x, double Q, double P2, int ip) {  
    vector<double> r(13);
    fevolvepdfp(&x, &Q, &P2, &ip, &r[0]);
    return r;
  }
  vector<double> xfxp(int nset, double x, double Q, double P2, int ip) {  
    vector<double> r(13);
    fevolvepdfpm(&nset, &x, &Q, &P2,&ip, &r[0]);
    return r;
  }

  void xfxp(double x, double Q, double P2, int ip, double* results) {
    fevolvepdfp(&x, &Q, &P2, &ip, results);
  }
  void xfxp(int nset, double x, double Q, double P2, int ip, double* results) {
    fevolvepdfpm(&nset, &x, &Q, &P2, &ip, results);
  }

  double xfxp(double x, double Q, double P2, int ip, int fl) {  
    vector<double> r(13);
    fevolvepdfp(&x, &Q, &P2, &ip, &r[0]);
    return r[fl+6];
  }
  double xfxp(int nset, double x, double Q, double P2, int ip, int fl) {  
    vector<double> r(13);
    fevolvepdfpm(&nset, &x, &Q, &P2, &ip, &r[0]);
    return r[fl+6];
  }


  vector<double> xfxa(double x, double Q, double a) {  
    vector<double> r(13);
    fevolvepdfa(&x, &Q, &a, &r[0]);
    return r;
  }
  vector<double> xfxa(int nset, double x, double Q, double a) {  
    vector<double> r(13);
    fevolvepdfam(&nset, &x, &Q, &a, &r[0]);
    return r;
  }

  void xfxa(double x, double Q, double a, double* results) {
    fevolvepdfa(&x, &Q, &a, results);
  }
  void xfxa(int nset, double x, double Q, double a, double* results) {
    fevolvepdfam(&nset, &x, &Q, &a, results);
  }

  double xfxa(double x, double Q, double a, int fl) {  
    vector<double> r(13);
    fevolvepdfa(&x, &Q, &a, &r[0]);
    return r[fl+6];
  }
  double xfxa(int nset, double x, double Q, double a, int fl) {  
    vector<double> r(13);
    fevolvepdfam(&nset, &x, &Q, &a, &r[0]);
    return r[fl+6];
  }


  vector<double> xfxphoton(double x, double Q) {  
    vector<double> r(13);  
    double mphoton;
    fevolvepdfphoton(&x, &Q, &r[0], &mphoton );
    r.push_back(mphoton);
    return r;
  }
  vector<double> xfxphoton(int nset, double x, double Q) {  
    vector<double> r(13);
    double mphoton;
    fevolvepdfphotonm(&nset, &x, &Q, &r[0], &mphoton );
    r.push_back(mphoton);
    return r;
  }

  void xfxphoton(double x, double Q, double* results) {
    fevolvepdfphoton(&x, &Q, results, results+13);
  }
  void xfxphoton(int nset, double x, double Q, double* results) {
    fevolvepdfphotonm(&nset, &x, &Q, results, results+13);
  }

  double xfxphoton(double x, double Q, int fl) {  
    vector<double> r(13);
    double mphoton;
    fevolvepdfphoton(&x, &Q, &r[0], &mphoton);
    if (fl == 7) return mphoton;
    return r[fl+6];
  }
  double xfxphoton(int nset, double x, double Q, int fl) {  
    vector<double> r(13);
    double mphoton;
    fevolvepdfphotonm(&nset, &x, &Q, &r[0], &mphoton);
    if ( fl == 7 ) return mphoton;
    return r[fl+6];
  }


  /////////////////////


  void getDescription() {
    fgetdesc();
  }
  void getDescription(int nset) {
    fgetdescm(&nset);
  }

  int numberPDF() {
    int N;
    fnumberpdf(&N); 
    return N;
  }
  int numberPDF(int nset) {
    int N; 
    fnumberpdfm(&nset, &N); 
    return N;
  }

  double alphasPDF(double Q) {
    double a;
    falphaspdf(&Q, &a);
    return a;
  }
  double alphasPDF(int nset, double Q) {
    double a;
    falphaspdfm(&nset, &Q, &a);
    return a;
  }

  int getOrderPDF() {
    int N;
    fgetorderpdf(&N);
    return N;
  }
  int getOrderPDF(int nset) {
    int N;
    fgetorderpdfm(&nset, &N);
    return N;
  }

  int getOrderAlphaS() {
    int N;
    fgetorderas(&N);
    return N;
  }
  int getOrderAlphaS(int nset) {
    int N;
    fgetorderasm(&nset, &N);
    return N;
  }

  double getQMass(int f) {
    double m;
    fgetqmass(&f, &m);
    return m;
  }
  double getQMass(int nset, int f) {
    double m;
    fgetqmassm(&nset, &f, &m);
    return m;
  }

  double getThreshold(int f) {
    double m;
    fgetthreshold(&f, &m);
    return m;
  }
  double getThreshold(int nset, int f) {
    double m;
    fgetthresholdm(&nset, &f, &m);
    return m;
  }

  int getNf() {
    int N; 
    fgetnf(&N); 
    return N;
  }
  int getNf(int nset) {
    int N; 
    fgetnfm(&nset, &N); 
    return N;
  }

  double getLam4(int m) {
    double l;
    fgetlam4(&m, &l);
    return l;
  }
  double getLam4(int nset, int m) {
    double l;
    fgetlam4m(&nset, &m, &l);
    return l;
  }

  double getLam5(int m) {
    double l;
    fgetlam5(&m, &l);
    return l;
  }
  double getLam5(int nset, int m) {
    double l;
    fgetlam5m(&nset, &m, &l);
    return l;
  }

  double getXmin(int m) {
    double xmin;
    fgetxmin(&m, &xmin);
    return xmin;
  }
  double getXmin(int nset,int m) {
    double xmin;
    fgetxminm(&nset, &m, &xmin);
    return xmin;
  }

  double getXmax(int m) {
    double xmax;
    fgetxmax(&m, &xmax);
    return xmax;
  }
  double getXmax(int nset,int m) {
    double xmax;
    fgetxmaxm(&nset, &m, &xmax);
    return xmax;
  }

  double getQ2min(int m) {
    double q2min;
    fgetq2min(&m, &q2min);
    return q2min;
  }
  double getQ2min(int nset,int m) {
    double q2min;
    fgetq2minm(&nset, &m, &q2min);
    return q2min;
  }

  double getQ2max(int m) {
    double q2max;
    fgetq2max(&m, &q2max);
    return q2max;
  }

  void extrapolate(bool extrapolate) { 
    if (extrapolate) {
      fextrapolateon();
    } else {
      fextrapolateoff();
    }
  }
  double getQ2max(int nset,int m) {
    double q2max;
    fgetq2maxm(&nset, &m, &q2max);
    return q2max;
  }


}
